%{

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>

#include "info2html.h"

#define MENU_START  "<table>"
#define MENU_ITEM   "<br><tr><td>"
#define MENU_DESC   " <td>"
#define MENU_END    "</table><p>"

%}

WORDS      [^:\n]*

SHORTLINK  "* "{WORDS}"::"
LONGLINK   "* "{WORDS}": "(\({WORDS}\))?{WORDS}\.
NOTELINK1  \*[Nn]ote[\040\n]{WORDS}"::"
NOTELINK2  \*[Nn]ote[\040\n]([^:]*)": "(\({WORDS}\))?([^\.\,]*)[\.\,]

UNDERLINE  ((={3,})|(-{3,})|(\*{3,})|(\.{3,}))
HEADLINE   ^.*\n/{UNDERLINE}\n

NUMLIST    \040{1,3}([0-9]){1,2}(\.|\))\040
ITEMLIST   \040{2,4}[\*]\040
DESCLIST1  \040{1,3}("-"){1,2}\040
DESCLIST2  "`".*"'"\n

%x menu nlist ilist dlist1 dlist2

%%

  /* kick out (aeh, no link) the "* Menu:" line */
<INITIAL>^"* Menu:".*      printf("<p><b>%s</b>",yytext);
  /* new paragraph */
<INITIAL>^\n               printf("<p>\n");
  /* headline */
<INITIAL>{HEADLINE}        printf("<h3>%s</h3>",yytext);
<INITIAL>^{UNDERLINE}\n    /* eat */

  /* menu & index */
<INITIAL>^{LONGLINK}       {
                             BEGIN(menu);printf(MENU_START);
			     printf(MENU_ITEM);
			     print_longlink(yytext);
			   }
<INITIAL>^{SHORTLINK}      {
                             BEGIN(menu);printf(MENU_START);
			     printf(MENU_ITEM);
			     print_shortlink(yytext);
			     printf(MENU_DESC);
			   }
<menu>^{LONGLINK}          {
			     printf(MENU_ITEM);
			     print_longlink(yytext);
			   }
<menu>^{SHORTLINK}         {
			     printf(MENU_ITEM);
			     print_shortlink(yytext);
			     printf(MENU_DESC);
			   }
<menu>^\n{LONGLINK}        {
			     printf(MENU_ITEM);
			     print_longlink(yytext+1);
			   }
<menu>^\n{SHORTLINK}       {
			     printf(MENU_ITEM);
			     print_shortlink(yytext+1);
			     printf(MENU_DESC);
			   }
<menu>^\n                  printf(MENU_END); BEGIN(INITIAL);

  /* lists */
<INITIAL>^{NUMLIST}        BEGIN(nlist); printf("<ol><li>");
<nlist>^\n{NUMLIST}        printf("<p><li>");
<nlist>^\n\040{4,}         printf("<p>");
<nlist>^\n                 printf("</ol>"); BEGIN(INITIAL);

<INITIAL>^{ITEMLIST}       BEGIN(ilist); printf("<ul><li>");
<ilist>^\n{ITEMLIST}       printf("<p><li>");
<ilist>^\n\040{4,}         printf("<p>");
<ilist>^\n                 printf("</ul>"); BEGIN(INITIAL);

<INITIAL>^{DESCLIST1}      BEGIN(dlist1); printf("<dl><dt>-- ");
<INITIAL>^{DESCLIST2}      BEGIN(dlist2); printf("<dl><dt>%s<dd>",yytext);
<dlist2>^(\n)?{DESCLIST2}  printf("<dt>%s<dd>",yytext);
<dlist1>\n(\040){6,}       /* eat (line break) */
<dlist1>\n                 BEGIN(dlist2); printf("<dd>");
<dlist2>^(\n)?{DESCLIST1}  BEGIN(dlist1); printf("<dt>--");
<dlist2>^\n\040{4,}        printf("<p>");
<dlist2>^\n                printf("</dl>"); BEGIN(INITIAL);

  /* preformated stuff */
<INITIAL>^\040{4,} |
<nlist,ilist>^\040{8,} |
<dlist2>^\040{8,}          {
                             char c;
			    
			     printf("<pre>%s",yytext);
			     for (;;) {
			       c = input();
			       putchar(c);
			       if (c == '\n') {
				 c = input();
				 if (c == '\n') {
				   unput('\n');
				   break;
				 } else {
				   putchar(c);
				 }
			       }
			     }
			     printf("</pre>");
			   }

  /* links in text (*note ...::) */
<INITIAL,nlist,ilist,dlist2>{NOTELINK1}  print_shortlink(yytext);
<INITIAL,nlist,ilist,dlist2>{NOTELINK2}  print_longlink(yytext);

  /* special handling for some chars */
<*>"<"                     printf("&lt;");
<*>">"                     printf("&gt;");
<*>"&"                     printf("&amp;");

  /* Ende */
<<EOF>> |
<*>^\x1f                   {
                             if (YY_START == menu)   printf(MENU_END);
                             if (YY_START == nlist)  printf("</ol>");
                             if (YY_START == ilist)  printf("</ul>");
			     if (YY_START == dlist1) printf("<dd></dl>");
			     if (YY_START == dlist2) printf("</dl>");
			     return 0;
			   }
%%

void print_longlink(char *link)
{
  char linknode[LINK_LENGTH];
  char linkfile[LINK_LENGTH];
  char text[LINK_LENGTH];
  char *h;

  /* text */
  strcpy(text,link+(link[1]==' ' ? 2 : 6));
  *(strchr(text,':')) = '\0';

  /* link */
  h = strchr(link,':');
  do h++; while (*h == ' ');
  if (*h == '(') {
    /* we have a file */
    strcpy(linkfile,h+1);
    *(h=strchr(linkfile,')')) = '\0';
    do h++; while (*h == ' ');
    strcpy(linknode,"Top"); /* default node, may be overwritten later */
  } else {
    strcpy(linkfile,file);
  }
  if (*h != '.' && *h != ',') {
    /* we have a node */
    quote(linknode,h);
    h = strstr(linknode,"%2e"); /* quoted '.' */
    if (!h)
      h = strstr(linknode,"%2c"); /* quoted ',' */
    *h = '\0';
  }
  printf("<a href=\"%s/%s/%s\">*",whoami,linkfile,linknode);
  if (link[1] != ' ') printf("note");
  printf(" %s:</a>",text);
  if (link[1] == ' ') printf(MENU_DESC);
  h = strchr(link,':');
  do h++; while (*h == ' ');
  printf("%s",h); 
}

void print_shortlink(char *link)
{
  char linknode[LINK_LENGTH];

  quote(linknode,link+(link[1]==' ' ? 2 : 6));
  *(strstr(linknode,"%3a")) = '\0'; /* ':' -> "%3a" (URL-Quoting) */

  printf("<a href=\"%s/%s/%s\">",whoami,file,linknode);
  printf("%s</a>",link); 
}
