/* needed to get the strdup prototype (linux) */
#define _BSD_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/stat.h>
#include <sys/time.h>

#include "info2html.h"

#define ERROR_1 "<html><head><title>ERROR</title></head><body>"
#define ERROR_2 "</body></html>"

/*-------------------------------------------------------------*/

char file[LINK_LENGTH];
char node[LINK_LENGTH];
char whoami[256];

/*-------------------------------------------------------------*/

void quote(char *dest, char *src)
{
  static const char lookup[16] = "0123456789abcdef";
  int  i,j;
  
  for (i=0, j=0; src[j] != '\0';) {
    if (isalnum(src[j])) {
      dest[i++] = src[j++];
    } else {
      dest[i++] = '%';
      dest[i++] = lookup[src[j]/16];
      dest[i++] = lookup[src[j]%16];
      j++;
    }
  }
  dest[i] = '\0';
}

/* copy string and convert the "%??" to normal char's */
void unquote(char *dest, char *src)
{
  int  i,j,hex;
  char nr[3];
  
  for (i=0, j=0; src[j] != '?' && src[j] != '\0'; i++) {
    if (src[j] == '%') {
      nr[0] = src[j+1];
      nr[1] = src[j+2];
      nr[2] = 0;
      sscanf(nr,"%x",&hex);
      dest[i]=hex;
      j+=3;
    } else {
      dest[i] = src[j];
      j+=1;
    }
  }
  dest[i] = '\0';
}

/*-------------------------------------------------------------*/

/* read the PATH_INFO enviroment variable and fill file+node */
void read_pathinfo(void)
{
  char *path,*h;

  file[0] = 0;
  node[0] = 0;
  path = getenv("PATH_INFO");
  if (path && *path) {
    strcpy(file,path+1);
    h = strchr(file,'/');
    if (h) {
      *(h++) = 0;
      if (*h) strcpy(node,h);
      /* workaround: Spinner appends a slash to the URL */
      if (NULL != (h = strchr(node,'/'))) *h = 0;
    }
  }
  if (!file[0]) strcpy(file,"dir");
  if (!node[0]) strcpy(node,"Top");
}

/* find our name */
void get_scriptname(char *argv0)
{
  char *h;

  /* first look for "SCRIPT_NAME"... */
  if (NULL == (h = getenv("SCRIPT_NAME"))) {
    /* ... if this fails, take basename of argv[0] */
    if (NULL == (h = strrchr(argv0,'/'))) {
      h = argv0;
    } else {
      h++;
    }
  }
  strcpy(whoami,h);
}

/*-------------------------------------------------------------*/

char infodirs[]=INFODIRS;
char *dirs[20];

void split_infodirs()
{
  char *h;
  int i;

  h = getenv("INFOPATH");
  if (h)
    dirs[0] = strdup(h);
  else
    dirs[0] = infodirs;

  for (i = 0; dirs[i] != NULL ; i++) {
    h = strchr(dirs[i],':');
    if (h) {
      *h = 0;
      dirs[i+1] = h+1;
    } else {
      dirs[i+1] = NULL;
    }
  }
}

/*-------------------------------------------------------------*/

/* open an infofile. The indir-parameter is for indirections */
FILE *open_infofile(int indir)
{
  char         filename[LINK_LENGTH],nr[10],last_modi[80];
  struct stat  st;
  int          fd_pipe[2];
  int          i,d,gzip;

  if (indir) {
    sprintf(nr,"-%i",indir);
  } else {
    nr[0] = 0;
  }
  for (d = 0; dirs[d] != NULL ; d++) {
    for (i = 0; i < 4 /* 8 */; i++) {
      switch (i) {
      case 0:
	gzip = FALSE;sprintf(filename,"%s/%s%s",dirs[d],file,nr);
	break;
      case 1:
	gzip = FALSE;sprintf(filename,"%s/%s.info%s",dirs[d],file,nr);
	break;
      case 2:
	gzip = TRUE; sprintf(filename,"%s/%s%s.gz",dirs[d],file,nr);
	break;
      case 3:
	gzip = TRUE; sprintf(filename,"%s/%s.info%s.gz",dirs[d],file,nr);
	break;
/*
      case 4:
	gzip = TRUE; sprintf(filename,"%s/%s%s.z",dirs[d],file,nr);
	break;
      case 5:
	gzip = TRUE; sprintf(filename,"%s/%s.info%s.z",dirs[d],file,nr);
	break;
      case 6:
	gzip = TRUE; sprintf(filename,"%s/%s%s.Z",dirs[d],file,nr);
	break;
      case 7:
	gzip = TRUE; sprintf(filename,"%s/%s.info%s.Z",dirs[d],file,nr);
	break;
*/
      default: return NULL; /* keep compiler happy */
      }
      if (-1!=stat(filename,&st) && S_ISREG(st.st_mode)) {
	if (!indir) {
	  /* http header */
	  puts("Content-type: text/html");
	  strftime(last_modi,79,"%a %b %d %H:%M:%S GMT %Y",
		   gmtime(&st.st_mtime));
	  printf("Last-modified: %s\n",last_modi);
	  puts("");
	}
	
	if (gzip) {
	  /* gzipped info-file */
	  pipe(fd_pipe);
	  switch (fork()) {
	  case -1:
	    return NULL;
	  case 0:
	    /* child */
	    dup2(fd_pipe[1],1);
	    close(fd_pipe[0]);
	    close(fd_pipe[1]);
	    execlp(GZIP,GZIP,"-cd",filename,NULL);
	    exit(1);
	  default:
	    /* parent */
	    dup2(fd_pipe[0],0);
	    close(fd_pipe[0]);
	    close(fd_pipe[1]);
	    return stdin;
	  }
	} else {
	  return fopen(filename,"r");
	}
      }
    }
  }
  return NULL;
}

/*-------------------------------------------------------------*/

void print_headlink(char *buf, char *what)
{
  char text[LINK_LENGTH];
  char linkfile[LINK_LENGTH];
  char linknode[LINK_LENGTH];
  char *h;

  h = strstr(buf,what);
  if (!h) return;
  h += strlen(what);
  strcpy(text,h);
  h = strchr(text,',' ); if (h) *h = '\0';
  h = strchr(text,'\t'); if (h) *h = '\0';
  h = strchr(text,'\n'); if (h) *h = '\0';
  if (text[0] == '(') {
    strcpy(linkfile,text+1);
    *(strchr(linkfile,')')) = '\0';
    strcpy(linknode,"Top");
  } else {
    strcpy(linkfile,file);
    quote(linknode,text);
  }
  printf("%s <a href=\"%s/%s/%s\">%s</a> \n",
	 what,whoami,linkfile,linknode,text);
}

/*-------------------------------------------------------------*/

void
log_error(int node_failed)
{
  time_t tm;
  char   timestr[64],from[256];

  time(&tm);
  strftime(timestr,64,"%a %h %d %T %Y",
	   localtime(&tm));
  if (getenv("HTTP_REFERER")) /* be sure we have a referer */
    sprintf(from," from %250s",getenv("HTTP_REFERER"));
  else
    from[0] = 0;
  if (node_failed)
    fprintf(stderr,"[%s] gnuinfo: access to node %s, info file %s failed%s\n",
	    timestr,node,file,from);
  else
    fprintf(stderr,"[%s] gnuinfo: access to info file %s failed%s\n",
	    timestr,file,from);
}

/*-------------------------------------------------------------*/

void find_node(int indir)
{
  int  indir_table[256];
  int  indirect = FALSE;
  char *this_node;
  char *h;
  int  i,line;
  FILE *fp;
  char buf[256];
  
  fp = open_infofile(indir);
  if (NULL == fp) {
    /* convert to lowercase and try again... */
    for (i = 0; file[i] != '\0'; i++)
      file[i] = tolower(file[i]);
    fp = open_infofile(indir);
    if (NULL == fp) {
      /* Hmm, no chance at all */
      log_error(0);
      printf("content-type: text/html\n\n");
      printf("%s\n<h1>ERROR</h1>can't open GNU info file for `<i>%s</i>'\n%s\n",
	     ERROR_1,file,ERROR_2);
      return;
    }
  }
  
  for (;;) {
    
    /* loop until EOF or ^? */
    do {
      if (NULL == fgets(buf,160,fp)) {
	log_error(1);
	printf("%s\n<h1>ERROR</h1>node `<i>%s</i>' not found in `<i>%s</i>'\n%s\n",
	       ERROR_1,node,file,ERROR_2);
	return;
      }
    } while (buf[0] != 31);
    
    /* read next line */
    fgets(buf,160,fp);
    
    /* check for indirect */
    if (!strncmp(buf,"Indirect:",9)) {
      indirect = TRUE;
      break;
    }

    /* compare node */
    this_node = strstr(buf,"Node: ");
    if (this_node) {
      this_node += 6;
      h = strchr(this_node,','); /* name may be terminated with a comma... */
      if (!h) h = strchr(this_node,'\t');  /* ... a tab ... */
      if (!h) h = strchr(this_node,'\n');  /* ... or a newline */
      if (h) *h = '\0';
      if (!strcasecmp(this_node,node)) {
	if (h) *h = ',';
	break;
      }
    }
  }

  if (indirect) {
    
    /* read indirect table */
    for (i=0;;) {
      fgets(buf,160,fp);
      if (buf[0] == 31)
	break;
      h = strchr(buf,':');
      if (h) {
	h++;
	indir_table[i++] = atoi(h);
      }
    }
    indir_table[i] = 0;
    
    /* read note table */
    for (;;) {
      if (NULL == fgets(buf,160,fp)) {
	log_error(1);
	printf("%s\n<h1>ERROR</h1>node `<i>%s</i>' not found in `<i>%s</i>'\n%s\n",
	       ERROR_1,node,file,ERROR_2);
	return;
      }
      this_node = strstr(buf,"Node: ");
      if (this_node) {
	this_node += 6;
	h = strchr(this_node,127);
	if (h) *h = '\0';
	if (!strcasecmp(this_node,node)) {
	  if (h) *h = 127;
	  break;
	}
      }
    }

    /* find out which file */
    h = strchr(buf,127); h++;
    line = atoi(h);
    for (i = 0; indir_table[i] != 0 && line >= indir_table[i] ; i++);

    /* play it again... */
    find_node(i);
  } else {

    /* html header */
    puts("<html>");
    puts("<head>");
    printf("<title>GNU info - %s - %s</title>\n",file,node);
    puts("</head>");
    puts("<body>");
    
    /* title */
    printf("<h1>%s - %s</h1>\n",file,node);
    h = strstr(buf,"File: ");
    if (h) {
      for (i=0;
	   h[i]!=','&&h[i]!='\t'&&h[i]!='\n'&&h[i]!='\0';
	   putchar(h[i++]));
      printf(" \n");
    }
    print_headlink(buf,"Node: ");
    print_headlink(buf,"Next: ");
    print_headlink(buf,"Prev: ");
    print_headlink(buf,"Up: ");
    printf("<hr>");

    /* go */
    yyin = fp;
    yylex();

    /* bottom */
    printf("<hr>");
    print_headlink(buf,"Next: ");
    print_headlink(buf,"Up: ");

    /* html footer */
    printf("<p><i>converted by <a href=\"http://www.cs.tu-berlin.de/~kraxel/info2html/gnuinfo.html\">%s</a></i>",whoami);
    puts("</body>");
    puts("</html>");
  }
}

/*-------------------------------------------------------------*/

int main(int argc, char *argv[])
{
  read_pathinfo();
  get_scriptname(argv[0]);
  split_infodirs();
  find_node(0);
  
  return(0);
}
